################################################################################
# cspyce/alias_support.py
# Used internally by cspyce; not intended for direct import.
################################################################################

import sys
import re
import inspect
import numbers
from collections import deque

import cspyce
import cspyce.cspyce1 as cspyce1

PYTHON2 = sys.version_info[0] < 3

# Global dictionaries used to track aliases
BODY_CODE_ALIASES = {}
BODY_NAME_ALIASES = {}
FRAME_CODE_ALIASES = {}
FRAME_NAME_ALIASES = {}
FRAME_CODE_OVERRIDES = {}
FRAME_NAME_OVERRIDES = {}

def get_body_aliases(item):
    """Return a tuple containing the list of equivalent codes and the list of
    equivalent names for a given body name or code."""

    item = _as_key(item)
    try:
        return (BODY_CODE_ALIASES[item], BODY_NAME_ALIASES[item])
    except KeyError:
        pass

    try:
        if isinstance(item, str):
            item = cspyce1.bodn2c_error(item)

        return ([item], [cspyce1.bodc2n_error(item)]) # this corrects name case
    except KeyError:
        pass

    try:
        if isinstance(item, str):
            item = int(item)

        return ([item], [cspyce1.bodc2n_error(item)])
    except (KeyError, ValueError):
        pass

    return ([], [])

def get_frame_aliases(item):
    """Return a tuple containing the list of equivalent codes and the list of
    equivalent names for a given frame name or code. If the argument
    identifies a body instead of a frame, then return the aliases of the
    associated frame."""

    item = _as_key(item)
    try:
        return (FRAME_CODE_ALIASES[item], FRAME_NAME_ALIASES[item])
    except KeyError:
        pass

    try:
        if isinstance(item, str):
            item = cspyce1.namfrm_error(item)

        return ([item], [cspyce1.frmnam_error(item)]) # this corrects name case
    except KeyError:
        pass

    try:
        if isinstance(item, str):
            item = int(item)

        return ([item], [cspyce1.frmnam_error(item)])
    except (KeyError, ValueError):
        pass

    # If it's not a frame, see if it's a body with an associated frame
    body_codes = get_body_aliases(item)[0]
    for body_code in body_codes:

        # If it's a spacecraft, the bus frame code is body code * 1000
        if body_code < 0:
            results = get_frame_aliases(body_code * 1000)
            if results[0]:
                return results

        # If it's a planetary body...
        else:
            (frame_code, frame_name, found) = cspyce.cidfrm.flag(body_code)
            if found:
                results = get_frame_aliases(frame_code)
                if results[0]:
                    return results

    return ([], [])

########################################

def define_body_aliases(*items):
    """Define a list of items (integer codes or strings) as aliases of the same
    body or spacecraft.

    The item list will be expanded to include other the names and codes already
    associated with each listed item. Higher-priority codes and names should
    come first. Note that name strings can be defined in mixed case, so they
    should be given exactly as you wish them to appear.
    """

    queue = deque(items) # Make a copy
    code_list = []
    name_list = []  # Stripped of surrounding whitespace; not otherwise modified
    aliases = set()

    # Expand lists to include everything
    while len(queue) > 0:
        item = queue.popleft()

        if item in aliases:
            continue

        (codes, names) = get_body_aliases(item)
        queue.extend(codes + names)

        if isinstance(item, numbers.Integral):
            code_list.append(item)
            aliases.add(int(item))

        else:
            key = _as_key(item)
            aliases.add(key)

            item = _clean_name(item)
            name_list.append(item)
            aliases.add(item)

    # Expand name list and generate associated keys (capitalized)
    name_list = _expand_name_list(code_list, name_list)
    key_list = [_as_key(n) for n in name_list]

    # Save lists in the global dictionary, keyed by each code, name and key
    for item in code_list + name_list + key_list:
        BODY_CODE_ALIASES[item] = code_list
        BODY_NAME_ALIASES[item] = name_list

    # Expand the code list length to match that of the name list, if necessary
    needed = len(name_list) - len(code_list)
    if needed > 0:
        code_list = code_list + needed * [code_list[-1]]

        # Move the name for this code to the end, so it takes precedence
        k = -needed - 1
        name_list = name_list[:k] + name_list[k+1:] + name_list[k:k+1]

    # Update the kernel pool
    for (name, code) in zip(name_list, code_list):

        # Don't enter info that is already there, because the name pool is small
        (test_code, found) = cspyce1.bodn2c.flag(name)
        if found and test_code == code:
            continue

        (test_name, found) = cspyce1.bodc2n.flag(code)
        if found and test_name.upper() == code:
            continue

        cspyce1.boddef(name, code)

def define_frame_aliases(*items):
    """Define a list of items (integer codes or strings) as aliases of the same
    coordinate frame.

    The item list will be expanded to include other the names and codes already
    associated with each listed item. Higher-priority codes and names should
    come first. Note that name strings can be defined in mixed case, so they
    should be given exactly as you wish them to appear.
    """

    queue = deque(items) # Make a copy
    code_list = []
    name_list = []  # Stripped of surrounding whitespace; not otherwise modified
    aliases = set()

    # Expand lists to include everything
    while len(queue) > 0:
        item = queue.popleft()

        if item in aliases:
            continue

        (codes, names) = get_frame_aliases(item)
        queue.extend(codes + names)

        if isinstance(item, numbers.Integral):
            code_list.append(item)
            aliases.add(int(item))

        else:
            key = _as_key(item)
            aliases.add(key)

            item = _clean_name(item)
            name_list.append(item)
            aliases.add(item)

    # Expand name list and generate associated keys (capitalized)
    name_list = _expand_name_list(code_list, name_list)
    key_list = [_as_key(n) for n in name_list]

    # Save lists in the global dictionary, keyed by each code, name and key
    for item in code_list + name_list + key_list:
        FRAME_CODE_ALIASES[item] = code_list
        FRAME_NAME_ALIASES[item] = name_list

    # Expand the code list length to match that of the name list, if necessary
    needed = len(name_list) - len(code_list)
    if needed > 0:
        code_list = code_list + needed * [code_list[-1]]

    # Otherwise, add these pairs to the dictionary of frame overrides`
    else:
        for (name, code) in zip(name_list, code_list):
            key = _as_key(name)
            cspyce1.FRAME_NAME_OVERRIDES[code] = name
            cspyce1.FRAME_NAME_OVERRIDES[name] = name
            cspyce1.FRAME_NAME_OVERRIDES[key ] = name
            cspyce1.FRAME_CODE_OVERRIDES[code] = code
            cspyce1.FRAME_CODE_OVERRIDES[name] = code
            cspyce1.FRAME_CODE_OVERRIDES[key ] = code

def _as_key(name):
    """Convert names to upper case and replace duplicated spaces with one;
    return ints unchanged.
    """

    if isinstance(name, numbers.Integral):
        return int(name)

    if isinstance(name, str):
        name = name.upper().strip()
        while ('  ' in name):
            name = name.replace('  ', ' ')

    return name

def _clean_name(name):
    """Strip surrounding whitespace."""

    return name.strip()

_VERSION_REGEX = re.compile('(.*) +V[0-9]+$')

def _expand_name_list(code_list, name_list):
    """Make sure there is at least one name for each code."""

    lcodes = len(code_list)
    if lcodes <= len(name_list):
        return name_list

    # Remove versioned names (ending with " Vn" for integer n)
    unversioned_names = []
    for name in name_list:
        matchobj = _VERSION_REGEX.match(name)
        if matchobj is None:
            unversioned_names.append(name)
        else:
            unversioned_names.append(matchobj.group(1))

    # Create multiple versions of lowest-priority name
    name_list = unversioned_names
    needed = lcodes - len(name_list)
    last_name = name_list[-1]
    for v in range(needed,0,-1):
        name_list.append(last_name + ' V' + str(v))

    return name_list

################################################################################
# cspyce alias function wrapper
################################################################################

ALIAS_NOTE = """
This version supports aliases, meaning that any input identifying a
SPICE body or frame can be specified using any of its aliases (integer
or name). The cspyce function will try each of the aliases in their order
of priority, and the first valid results will be returned.

**These inputs can be given as either a name or an integer code.
"""

def _alias_signature(func):
    """Add "**" to the end of body and frame input definitions."""

    for (sig, name) in zip(func.SIGNATURE, func.ARGNAMES):
        if sig in ('body_name', 'body_code', 'frame_name', 'frame_code'):
            definition = func.DEFINITIONS[name]
            if not definition[1].endswith('**'):
                func.DEFINITIONS[name] = (definition[0], definition[1] + '**')

def _alias_name(name):
    parts = name.split('_')
    if len(parts) > 1 and parts[1] == 'alias':
        return name

    parts = [parts[0], 'alias'] + parts[1:]
    return '_'.join(parts)

def _validate_args(func, args, keywords):
    """Signal error if an alias is of the wrong type."""

    for indx in list(range(len(args))) + list(keywords.keys()):

        if indx not in func.ALIAS_ARGS:
            continue

        # Get argument type and translate
        arg = _getarg(indx, args, keywords)
        argtype = func.ALIAS_ARGS[indx]

        if isinstance(arg, int) and not argtype[-4:] == 'code':
            name = indx if isinstance(indx, str) else func.ARGNAMES[indx]
            body_or_frame = argtype.split('_')[0]
            cspyce1.setmsg('%s code %d not found in kernel pool' %
                           (body_or_frame, arg))
            cspyce1.sigerr('SPICE(' + body_or_frame.upper() + 'IDNOTFOUND)')

        elif isinstance(arg, str) and not argtype[-4:] == 'name':
            name = indx if isinstance(indx, str) else func.ARGNAMES[indx]
            body_or_frame = argtype.split('_')[0]
            cspyce1.setmsg('%s name "%s" not found in kernel pool' %
                           (body_or_frame, arg))
            cspyce1.sigerr('SPICE(' + body_or_frame.upper() + 'NAMENOTFOUND)')

def alias_version(func):
    """Wrapper function to support aliasing of SPICE body names or codes."""

    if hasattr(func, 'alias'):
        return func.alias

    if not hasattr(func, 'ALIAS_ARGS'):
        alias_args = {}
        for k in range(len(func.ARGNAMES)):
            sig = func.SIGNATURE[k]
            if sig not in ('body_name', 'body_code',
                           'frame_name', 'frame_code'):
                continue

            alias_args[k] = sig
            alias_args[func.ARGNAMES[k]] = sig

        func.ALIAS_ARGS = alias_args

    if not func.ALIAS_ARGS:
        func.alias   = func
        func.noalias = func
        return func

    def wrapper(*args, **keywords):
        return _exec_with_aliases(wrapper, func, *args, **keywords)

    # Copy type info but not function version links
    for (key, value) in func.__dict__.items():
        if not callable(value):
            wrapper.__dict__[key] = value

    _alias_signature(wrapper)   # Add asterisks on aliased inputs

    # Save key attributes of the wrapper function before returning
    cspyce1.assign_docstring(wrapper, ALIAS_NOTE)
    wrapper.__name__ = _alias_name(func.__name__)
    if PYTHON2:
        wrapper.func_defaults = func.func_defaults
    else:
        wrapper.__defaults__ = func.__defaults__
        wrapper.__signature__ = inspect.signature(func)

    # Insert mutual links
    wrapper.alias   = wrapper
    wrapper.noalias = func

    func.alias   = wrapper
    func.noalias = func

    return wrapper

ALIAS_LOOKUP_DICT = {    # returns (function, index)
    'body_code' : (get_body_aliases , 0),
    'body_name' : (get_body_aliases , 1),
    'frame_code': (get_frame_aliases, 0),
    'frame_name': (get_frame_aliases, 1),
}

def _getarg(indx, args, keywords):
    if isinstance(indx, numbers.Integral):
        return args[indx]
    else:
        return keywords[indx]

def _setarg(indx, value, args, keywords):
    if isinstance(indx, numbers.Integral):
        args[indx] = value
    else:
        keywords[indx] = value

def _exec_with_aliases(wrapper, func, *args, **keywords):
    """Main function to handle a list of aliases in place of a single body name
    or code."""

    # Identify arguments that might have aliases
    alias_indices = []
    alias_args = list(args)     # Make copies
    alias_keywords = keywords.copy()
    for indx in list(range(len(args))) + list(keywords.keys()):

        if indx not in func.ALIAS_ARGS:
            continue

        # Get argument type and translate
        arg = _getarg(indx, alias_args, alias_keywords)

        argtype = func.ALIAS_ARGS[indx]
        (lookup_func, code_or_name) = ALIAS_LOOKUP_DICT[argtype]
        options = lookup_func(arg)[code_or_name]

        # Interpret list of options
        if len(options) == 0:
            value = arg
        elif len(options) == 1:
            value = options[0]
        else:
            value = options     # using aliases here
            alias_indices.append(indx)

        _setarg(indx, value, alias_args, alias_keywords)

    # Call function now if iteration is not needed
    if not alias_indices:
        cspyce.chkin(wrapper.__name__)
        _validate_args(func, alias_args, alias_keywords)
        result = func.__call__(*alias_args, **alias_keywords)
        cspyce.chkout(wrapper.__name__)
        return result

    # Iterate through the the last optional index; return if the call does not
    # fail.
    temp_args = list(alias_args)
    temp_keywords = alias_keywords.copy()
    wrapper.__name__ = _alias_name(func.__name__)

    indx = alias_indices[0]
    options = _getarg(indx, temp_args, temp_keywords)
    for option in options:
        _setarg(indx, option, temp_args, temp_keywords)

        try:
            results = _exec_with_one_alias(alias_indices[1:], wrapper, func,
                                           *temp_args, **temp_keywords)

            # On success, attach the value of the code or name that worked
            argname = func.ARGNAMES[indx]
            wrapper.__dict__[argname] = option
            return results

        except _AliasFailure:
            continue

    # Nothing worked. Raise error from a call using original inputs, after
    # translation to the correct type

    args = list(args)   # make a mutable copy
    keywords = keywords.copy()

    # Use each original argument unless it is the wrong type
    for indx in alias_indices:
        argtype = func.ALIAS_ARGS[indx]
        value = _getarg(indx, args, keywords)
        if isinstance(value, numbers.Integral) and argtype.endswith('_code'):
            continue
        if isinstance(value, str) and argtype.endswith('_name'):
            continue

        options = _getarg(indx, alias_args, alias_keywords)
        _setarg(indx, options[0], args, keywords)

    # Call the function
    cspyce.chkin(wrapper.__name__)
    _validate_args(func, args, keywords)
    result = func.__call__(*args, **keywords)
    cspyce.chkout(wrapper.__name__)
    return result

def _exec_with_one_alias(alias_indices, wrapper, func, *args, **keywords):
    """Recursive function to evaluate the function using multiple aliases.

    At each call, the first alias index is used, and this function is called
    called recursively with each optional value of that input argument. It
    repeats until a NotImpementedError (a type of error not used elsewhere) is
    NOT raised. The absence of this error condition indicates success.

    On success, it saves the aliased value that worked as an attribute of the
    wrapper function, and then returns the successful value.

    If the list of alias indices is empty, then recursion stops and this
    evaluates the function with the given set of alias values. On success, it
    returns the result. On failure, it raises a _AliasFailure, which will
    cause the "parent" of this function to try again its next possible value
    of the alias.
    """

    # Handle multiple aliased indices by recursion
    if alias_indices:

      indx = alias_indices[0]
      options = _getarg(indx, args, keywords)
      for option in options:
        _setarg(indx, option, args, keywords)

        try:
            result = _exec_with_one_alias(alias_indices[1:], wrapper, func,
                                          *args, **keywords)

            # On success, attach the value of the code or name that worked
            argname = func.ARGNAMES[indx]
            wrapper.__dict__[argname] = option
            return result

        except _AliasFailure:
            continue

    # Recursion is done, so execute the function
    try:
        cspyce.chkin(wrapper.__name__)
        results = func.error.__call__(*args, **keywords)
        if cspyce.failed():
            cspyce.reset()
            cspyce.chkout(wrapper.__name__)
            raise NotImplementedError()

    except Exception:
        raise _AliasFailure()

    # Make sure the results correspond to the proper version of the function
    if func is func.error:
        return results

    return func.__call__(*args, **keywords)

class _AliasFailure(Exception): # used for iterative attempts at a working alias
    pass

################################################################################
# Define the alias function selector
################################################################################

SPYCE_DICT = cspyce.__dict__

def use_aliases(*funcs):
    """Switch the listed functions or names of functions to use the "alias"
    version by default. This affects all versions of any given cspyce function.
    If the list is empty, apply this operation to all cspyce functions.
    """

    if funcs:
        cspyce.GLOBAL_STATUS.add('ALIASES')
        cspyce.GLOBAL_STATUS.discard('NOALIASES')
    else:
        cspyce.GLOBAL_STATUS.discard('ALIASES')
        cspyce.GLOBAL_STATUS.discard('NOALIASES')

    for name in cspyce._get_func_names(funcs, source=SPYCE_DICT):
        # <name>_alias must always point to the alias version
        # Only function names that don't explicitly specify this are modified to
        # point to the alias version.
        if 'alias' not in name:
            SPYCE_DICT[name] = SPYCE_DICT[name].alias

def use_noaliases(*funcs):
    """Switch the listed functions or names of functions to use the "noalias"
    version by default. This affects all versions of any given cspyce function.
    If the list is empty, apply this operation to all cspyce functions.
    """

    if funcs:
        cspyce.GLOBAL_STATUS.discard('ALIASES')
        cspyce.GLOBAL_STATUS.add('NOALIASES')
    else:
        cspyce.GLOBAL_STATUS.discard('ALIASES')
        cspyce.GLOBAL_STATUS.discard('NOALIASES')

    for name in cspyce._get_func_names(funcs, source=SPYCE_DICT):
        # <name>_alias must always point to the alias version
        # Only function names that don't explicitly specify this are modified to
        # point to the noalias version.
        if 'alias' not in name:
            SPYCE_DICT[name] = SPYCE_DICT[name].noalias

################################################################################
# Function to define alias versions and links for all cspyce functions
################################################################################

def _define_all_alias_versions():
    """Generate all missing alias functions and set the "alias" and "noalias"
    attribute for all cspyce functions.

    This routine can be run multiple times. At each run, it only creates
    whatever is missing.
    """

    # Create an alias version of each function that needs one
    alias_pairs = []
    funcs = cspyce.get_all_funcs(SPYCE_DICT).values()
    for func in funcs:
        if hasattr(func, 'alias'):
            continue

        alias_func = alias_version(func)
        if alias_func is func:          # function could not use aliases
            continue

        alias_name = alias_func.__name__
        SPYCE_DICT[alias_name] = alias_func

        alias_pairs.append((alias_func, func))

    # At this point, every alias function exists and every cspyce function has
    # "alias" and "noalias" attributes. However, links to other associated
    # cspyce functions still need to be filled in.

    for (afunc, func) in alias_pairs:
        _define_missing_versions_for_alias(afunc, func)

def _define_missing_versions_for_alias(afunc, func):
    """Complete the missing version attributes for a given alias function."""

    for (key, version) in func.__dict__.items():
        if not callable(version):
            continue

        if 'noalias' in key:
            afunc.__dict__[key] = version
        else:
            version_name = _alias_name(version.__name__)
            afunc.__dict__[key] = SPYCE_DICT[version_name]

################################################################################



################################################################################
# get_spice.py
################################################################################
#
# This module is responsible for downloading the CSPICE source files based on
# whatever hardware and software it is currently running on.
#
# The contents of the directories
#        <download>/cspice/src/cspice
#        <download>/include
# are put into the two directories
#        cspice/<os>-<arch>/src
#        cspice/<os>-<arch>/include
#
# This module should only be called by setup.py.
#
# This file owes a big debt to Dr. Andrew Annex and his file:
#      https://github.com/AndrewAnnex/SpiceyPy/blob/main/get_spice.py
# His version is much more ambitious than this and also compiles the files into
# a shared library.
################################################################################


from __future__ import absolute_import, with_statement

import os
import platform
import shutil
import subprocess
import sys
import tempfile
import time
from tarfile import TarFile
from zipfile import ZipFile

try:
    from pathlib import Path
except ImportError:
    subprocess.check_call([sys.executable, "-m", "pip", "install", "pathlib"])
    from pathlib import Path


CSPICE_DISTRIBUTIONS = {
    # system   arch        distribution name           extension
    # -------- ----------  -------------------------   ---------
    ("Darwin", "x86_64"): ("MacIntel_OSX_AppleC_64bit", "tar.Z"),
    ("Darwin", "arm64"): ("MacM1_OSX_clang_64bit", "tar.Z"),
    ("cygwin", "x86_64"): ("PC_Cygwin_GCC_64bit", "tar.Z"),
    ("FreeBSD", "x86_64"): ("PC_Linux_GCC_64bit", "tar.Z"),
    ("Linux", "x86_64"): ("PC_Linux_GCC_64bit", "tar.Z"),
    ("Windows", "x86_64"): ("PC_Windows_VisualC_64bit", "zip"),
    # This is just so I can easily test it in a docker image on my Mac
    ("Linux", "arm64"): ("PC_Linux_GCC_64bit", "tar.Z"),
}

ARCHITECTURE_TRANSLATOR = {"aarch64": "arm64", "AMD64": "x86_64"}

# versions
SPICE_VERSION = "N0067"


class GetCspice(object):
    def __init__(self):
        self.host_OS = platform.system()
        architecture = platform.machine()
        self.architecture = ARCHITECTURE_TRANSLATOR.get(architecture, architecture)

        # Check if platform is Unix-like OS or not
        self.is_unix = self.host_OS in ("Linux", "Darwin", "FreeBSD")
        # Get directory that this file is in.
        self.root_dir = str(Path(os.path.realpath(__file__)).parent)
        # Get directory into which we want to put all our files
        self.target_directory = os.path.join(
            self.root_dir, "cspice", self.host_OS + "-" + self.architecture)

    def download(self):
        if (Path(os.path.join(self.target_directory, "src")).is_dir() and
            Path(os.path.join(self.target_directory, "include")).is_dir()):
            return self.target_directory
        # Note: Once we toss 2.7 support, this can be rewritten as "with ...."
        tmpdir = tempfile.mkdtemp()
        try:
            self.download_cspice(destination=tmpdir)
            shutil.copytree(os.path.join(tmpdir, "cspice", "src", "cspice"),
                            os.path.join(self.target_directory, "src"))
            shutil.copytree(os.path.join(tmpdir, "cspice", "include"),
                            os.path.join(self.target_directory, "include"))
        finally:
            shutil.rmtree(tmpdir)
        return self.target_directory

    def download_cspice(self, destination):
        try:
            # Get the remote file path for the Python architecture that
            # executes the script.
            assert sys.maxsize > 2 ** 32, "Machine must support 64bit"
            system = self.host_OS
            system = "cygwin" if "CYGWIN" in system else system
            distribution, extension = CSPICE_DISTRIBUTIONS[(system, self.architecture)]
        except KeyError:
            raise RuntimeError("CSpice does not support your system")

        cspice_filename = "cspice.{}".format(extension)
        url = ("https://naif.jpl.nasa.gov/pub/naif/misc/toolkit_{0}/C/{1}/packages/{2}"
               .format(SPICE_VERSION, distribution, cspice_filename))

        # Download the file
        target_file = os.path.join(destination, cspice_filename)
        attempts = 10  # Let's try a maximum of attempts for getting SPICE
        while attempts:
            try:
                if extension == "zip":
                    # TODO:
                    # Do we want --connect_timeout? This curl either seems to take
                    # 3s or several minutes, somewhat randomly.
                    subprocess.check_call(["curl", url, "-o", target_file])
                    with ZipFile(target_file, "r") as archive:
                        archive.extractall(destination)
                else:
                    target_file = target_file[:-2] # remove the .Z
                    subprocess.check_call("curl {} | gzip -d > {}".format(
                        url, target_file), shell=True)
                    with TarFile.open(target_file, "r") as archive:
                        archive.extractall(destination)
                os.unlink(target_file)
                break
            except (RuntimeError, subprocess.CalledProcessError) as error:
                attempts -= 1
                if attempts == 0:
                    raise error
                print("Download failed with URLError: {0}, trying again after "
                      "15 seconds!".format(error))
                time.sleep(15)


if __name__ == "__main__":
    GetCspice().download()



################################################################################
# cspyce/cspyce1.py
#
# This module provides several enhancements over the low-level cspyce0 interface
# to the CSPICE library. See cspyce/__init__.py for a full explanation.
#
# Used internally by cspyce; not intended for direct import.
################################################################################

from __future__ import print_function

import sys
import numpy as np
import textwrap

from cspyce import cspyce0
from cspyce.cspyce0 import *

from cspyce.cspyce1_info import \
    CSPYCE_SIGNATURES, CSPYCE_ARGNAMES, CSPYCE_DEFAULTS, \
    CSPYCE_RETURNS, CSPYCE_RETNAMES, CSPYCE_DEFINITIONS, \
    CSPYCE_ABSTRACT, CSPYCE_PS, CSPYCE_URL

# Global variables used below
import __main__
INTERACTIVE = not hasattr(__main__, '__file__')
PYTHON2 = sys.version_info[0] < 3

################################################################################
# GET/SET handling
################################################################################

def erract(op='', action=''):
    """Allow special argument handling:
        erract()            -> erract('GET', '')
        erract('GET')       -> erract('GET', '')
        erract('EXCEPTION') -> erract('SET', 'EXCEPTION')
    Also override 'ABORT' and 'DEFAULT' options in interactive mode.
    """

    op = op.upper().strip()
    if not op:
        op = 'GET'

    if op == 'GET':
        return cspyce0.erract('GET', '')

    if op != 'SET' and not action:  # single input value, assume 'SET'
        cspyce0.erract('SET', op)
        return op

    action = action.upper().strip()
    if op == 'SET' and INTERACTIVE and action in ('ABORT', 'DEFAULT'):
        print('ERROR action "{}" is disabled in interactive mode; '
              'using "RETURN"'.format(action))
        cspyce0.erract('SET', 'RETURN')
        return 'RETURN'

    return cspyce0.erract(op, action)

def errdev(op='', action=''):
    """Allow special argument handling:
        errdev()            -> errdev('GET', '')
        errdev('GET')       -> errdev('GET', '')
        errdev('SCREEN')    -> errdev('SET', 'SCREEN')
    """

    if not op:
        return cspyce0.errdev('GET', '')

    op_upper = op.upper()
    if op_upper == 'GET':
        return cspyce0.errdev('GET', '')

    if op_upper != 'SET' and not action:  # single input value, assume 'SET'
        cspyce0.errdev('SET', op)
        return op

    return cspyce0.errdev(op, action)

def errprt(op='', list_=''):
    """Allow special argument handling:
        errprt()            -> errprt('GET', '')
        errprt('GET')       -> errprt('GET', '')
        errprt('LONG')      -> errprt('SET', 'LONG')
    """

    op = op.upper().strip()
    if not op:
        op = 'GET'

    if op == 'GET':
        return cspyce0.errprt('GET', '')

    if op != 'SET' and not list_:  # single input value, assume 'SET'
        cspyce0.errprt('SET', op)
        return op

    list_ = list_.upper().strip()
    return cspyce0.errprt(op, list_)

################################################################################
# Define _error versions of functions that raise error conditions rather than
# return status flags.
#
# The code is written to work regardless of the type of error handling in use.
################################################################################

#### defined in cspyce0.i

def bodc2n_error(code):
    (name, found) = cspyce0.bodc2n(code)
    if not found:
        chkin('bodc2n_error')
        setmsg('body code {} not found in kernel pool'.format(code))
        sigerr('SPICE(BODYIDNOTFOUND)')
        chkout('bodc2n_error')

    return name

def bodn2c_error(name):
    (code, found) = cspyce0.bodn2c(name)
    if not found:
        chkin('bodn2c_error')
        setmsg('body name "{}" not found in kernel pool'.format(name))
        sigerr('SPICE(BODYNAMENOTFOUND)')
        chkout('bodn2c_error')

    return code

def bods2c_error(name):
    (code, found) = cspyce0.bods2c(name)
    if not found:
        chkin('bods2c_error')
        setmsg('body name "{}" not found in kernel pool'.format(name))
        sigerr('SPICE(BODYNAMENOTFOUND)')
        chkout('bods2c_error')

    return code

def ccifrm_error(frclss, clssid):
    (frcode, frname, center, found) = cspyce0.ccifrm(frclss, clssid)
    if not found:
        chkin('ccifrm_error')
        setmsg('unrecognized frame description: class {}; '
               'class id {}'.format(frclss, clssid))
        sigerr('SPICE(INVALIDFRAMEDEF)')
        chkout('ccifrm_error')

    return [frcode, frname, center]

def cidfrm_error(code):
    (frcode, name, found) = cspyce0.cidfrm(code)
    if not found:
        chkin('cidfrm_error')
        setmsg('body code {} not found in kernel pool'.format(code))
        sigerr('SPICE(BODYIDNOTFOUND)')
        chkout('cidfrm_error')

    return [frcode, name]

def ckcov_error(ck, idcode, needav, level, tol, timsys):
    coverage = cspyce0.ckcov(ck, idcode, needav, level, tol, timsys)
    if coverage.size == 0:
        chkin('ckcov_error')
        setmsg('body code {} not found in C kernel file {}'.format(idcode, ck))
        sigerr('SPICE(BODYIDNOTFOUND)')
        chkout('ckcov_error')

    return coverage

def ckgp_error(inst, sclkdp, tol, ref):
    (cmat, clkout, found) = cspyce0.ckgp(inst, sclkdp, tol, ref)
    if not found:
        name = cspyce0.frmnam(inst)
        if name:
            namestr = ' (' + name + ')'
        else:
            namestr = ''

        chkin('ckgp_error')
        setmsg('insufficient C kernel data to evaluate '
               'instrument/spacecraft {}{} '
               'at spacecraft clock time {} '
               'with tolerance {}'.format(inst, namestr, sclkdp, tol))
        sigerr('SPICE(CKINSUFFDATA)')
        chkout('ckgp_error')

    return [cmat, clkout]

def ckgpav_error(inst, sclkdp, tol, ref):
    (cmat, av, clkout, found) = cspyce0.ckgpav(inst, sclkdp, tol, ref)
    if not found:
        name = cspyce0.frmnam(inst)
        if name:
            namestr = ' (' + name + ')'
        else:
            namestr = ''

        chkin('ckgpav_error')
        setmsg('insufficient C kernel data to evaluate '
               'instrument/spacecraft {}{} '
               'at spacecraft clock time {} '
               'with tolerance {}'.format(inst, namestr, sclkdp, tol))
        sigerr('SPICE(CKINSUFFDATA)')
        chkout('ckgpav_error')

    return [cmat, av, clkout]

def cnmfrm_error(cname):
    (frcode, frname, found) = cspyce0.cnmfrm(cname)
    if not found:
        chkin('cnmfrm_error')
        setmsg('body name "{}" not found in kernel pool'.format(cname))
        sigerr('SPICE(BODYNAMENOTFOUND)')
        chkout('cnmfrm_error')

    return [frcode, frname]

def dtpool_error(name):
    (found, n, vtype) = cspyce0.dtpool(name)
    if not found:
        chkin('dtpool_error')
        setmsg('pool variable "{}" not found'.format(name))
        sigerr('SPICE(VARIABLENOTFOUND)')
        chkout('dtpool_error')

    return [n, vtype]

def frinfo_error(frcode):
    (cent, frclss, clssid, found) = cspyce0.frinfo(frcode)
    if not found:
        chkin('frinfo_error')
        setmsg('frame code {} not found in kernel pool'.format(frcode))
        sigerr('SPICE(FRAMEIDNOTFOUND)')
        chkout('frinfo_error')

    return [cent, frclss, clssid]

def frmnam1_error(frcode):  # change of name; frmnam_error is defined below
    frname = cspyce0.frmnam(frcode)
    if frname == '':
        chkin('frmnam_error')
        setmsg('frame code {} not found'.format(frcode))
        sigerr('SPICE(FRAMEIDNOTFOUND)')
        chkout('frmnam_error')

    return frname

def gcpool_error(name, start=0):
    (cvals, found) = cspyce0.gcpool(name, start)
    if not found:
        ok, _count, _nctype = cspyce0.dtpool(name)
        if not ok:
            chkin('gcpool_error')
            setmsg('pool variable "{}" not found'.format(name))
            sigerr('SPICE(VARIABLENOTFOUND)')
            chkout('gcpool_error')
            return []

    _ok, count, nctype = cspyce0.dtpool(name)
    if nctype != 'C':
        chkin('gcpool_error')
        setmsg('string information not available; '
               'kernel pool variable "{}" has numeric values'.format(name))
        sigerr('SPICE(WRONGDATATYPE)')
        chkout('gcpool_error')
        return []

    if start > count:
        chkin('gcpool_error')
        setmsg('kernel pool has only {} '
               'values for variable "{}";'
               'start index value {} is too large'.format(count, name, start))
        sigerr('SPICE(INDEXOUTOFRANGE)')
        chkout('gcpool_error')
        return []

    return cvals

def gdpool_error(name, start=0):
    (values, found) = cspyce0.gdpool(name, start)
    if not found:
        (ok, count, nctype) = cspyce0.dtpool(name)
        if not ok:
            chkin('gdpool_error')
            setmsg('pool variable "{}" not found'.format(name))
            sigerr('SPICE(VARIABLENOTFOUND)')
            chkout('gdpool_error')
            return []

    (ok, count, nctype) = cspyce0.dtpool(name)
    if nctype != 'N':
        chkin('gdpool_error')
        setmsg('numeric values are not available; '
               'kernel pool variable "{}" has string values'.format(name))
        sigerr('SPICE(WRONGDATATYPE)')
        chkout('gdpool_error')
        return []

    if start > count:
        chkin('gdpool_error')
        setmsg('kernel pool has only {} '
               'values for variable "{}";'
               'start index value {} is too large'.format(count, name, start))
        sigerr('SPICE(INDEXOUTOFRANGE)')
        chkout('gdpool_error')
        return []

    return values

def gipool_error(name, start=0):
    (ivals, found) = cspyce0.gipool(name, start)
    if not found:
        (ok, count, nctype) = cspyce0.dtpool(name)
        if not ok:
            chkin('gipool_error')
            setmsg('pool variable "{}" not found'.format(name))
            sigerr('SPICE(VARIABLENOTFOUND)')
            chkout('gipool_error')
            return []

    (ok, count, nctype) = cspyce0.dtpool(name)
    if nctype != 'N':
        chkin('gipool_error')
        setmsg('numeric values are not available; '
               'kernel pool variable "{}" has string values'.format(name))
        sigerr('SPICE(WRONGDATATYPE)')
        chkout('gipool_error')
        return []

    if start > count:
        chkin('gipool_error')
        setmsg('kernel pool has only {} '
               'values for variable "{}";'
               'start index value {} is too large'.format(count, name, start))
        sigerr('SPICE(INDEXOUTOFRANGE)')
        chkout('gipool_error')
        return []

    return ivals

def gnpool_error(name, start=0):
    (kvars, found) = cspyce0.gnpool(name, 0)
    if not found:
        chkin('gnpool_error')
        setmsg('no kernel pool variables found matching template "{}"'.format(name))
        sigerr('SPICE(VARIABLENOTFOUND)')
        chkout('gnpool_error')
        return []

    if start > len(kvars):
        setmsg('kernel pool has only {} '
               'variables matching template "{}";'
               'start index value {} is too large'.format(len(kvars), name, start))
        sigerr('SPICE(INDEXOUTOFRANGE)')
        chkout('gnpool_error')
        return []

    return kvars[start:]

def namfrm1_error(frname):  # change of name; namfrm_error is defined below
    frcode = cspyce0.namfrm(frname)
    if frcode == 0:
        chkin('namfrm_error')
        setmsg('frame name "{}" not found in kernel pool'.format(frname))
        sigerr('SPICE(FRAMENAMENOTFOUND)')
        chkout('namfrm_error')

    return frcode

def pckcov_error(pck, code):
    coverage = cspyce0.pckcov(pck, code)
    if coverage.size == 0:
        chkin('pckcov_error')
        setmsg('frame code {} not found in binary PC kernel file {}'.format(code, pck))
        sigerr('SPICE(FRAMEIDNOTFOUND)')
        chkout('pckcov_error')

    return coverage

def spkcov_error(spk, code):
    coverage = cspyce0.spkcov(spk, code)
    if coverage.size == 0:
        chkin('spkcov_error')
        setmsg('body code {} not found in SP kernel file {}'.format(code, spk))
        sigerr('SPICE(BODYIDNOTFOUND)')
        chkout('spkcov_error')

    return coverage

def srfc2s_error(code, bodyid):
    (srfstr, isname) = cspyce0.srfc2s(code, bodyid)
    if not isname:
        chkin('srfc2s_error')
        setmsg('surface for {}/{} not found'.format(code, bodyid))
        sigerr('SPICE(NOTRANSLATION)')
        chkout('srfc2s_error')

    return srfstr

def srfcss_error(code, bodstr):
    (srfstr, isname) = cspyce0.srfcss(code, bodstr)
    if not isname:
        chkin('srfcss_error')
        setmsg('surface for {}/"{}" not found'.format(code, bodstr))
        sigerr('SPICE(NOTRANSLATION)')
        chkout('srfcss_error')

    return srfstr

def srfs2c_error(srfstr, bodstr):
    (code, found) = cspyce0.srfs2c(srfstr, bodstr)
    if not found:
        chkin('srfs2c_error')
        setmsg('surface for "{}"/"{}" not found'.format(srfstr, bodstr))
        sigerr('SPICE(NOTRANSLATION)')
        chkout('srfs2c_error')

    return code

def srfscc_error(srfstr, bodyid):
    (code, found) = cspyce0.srfscc(srfstr, bodyid)
    if not found:
        chkin('srfscc_error')
        setmsg('"{}"/{} not found'.format(srfstr, bodyid))
        sigerr('SPICE(NOTRANSLATION)')
        chkout('srfscc_error')

    return code

def stpool_error(item, nth, contin):
    (string, found) = cspyce0.stpool(item, nth, contin)
    if not found:
        (ok, count, nctype) = cspyce0.dtpool(name)
        if not ok:
            chkin('stpool_error')
            setmsg('pool variable "{}" not found'.format(name))
            sigerr('SPICE(VARIABLENOTFOUND)')
            chkout('stpool_error')
            return ''

        if nth != 0:
            (_, ok) = cspyce0.stpool(item, 0, contin)
            if ok:
                chkin('stpool_error')
                setmsg('index too large; '
                       'kernel pool has fewer than {} '
                       'strings matching name "{}" '
                       'and continuation "{}"'.format(nth, item, contin))
                sigerr('SPICE(INDEXOUTOFRANGE)')
                chkout('stpool_error')
                return ''

    (ok, count, nctype) = cspyce0.dtpool(item)
    if nctype != 'C':
        setmsg('string values are not available; '
               'kernel pool variable "{}" has numeric values'.format(item))
        sigerr('SPICE(WRONGDATATYPE)')
        chkout('stpool_error')
        return ''

    return string

def tparse_error(string):
    (sp2000, msg) = cspyce0.tparse(string)
    if msg:
        chkin('tparse_error')
        setmsg(msg)
        sigerr('SPICE(INVALIDTIMESTRING)')
        chkout('tparse_error')

    return sp2000

def tpictr_error(string):
    (pictur, ok, msg) = cspyce0.tpictr(string)
    if not ok:
        chkin('tpictr_error')
        setmsg(msg)
        sigerr('SPICE(INVALIDTIMESTRING)')
        chkout('tpictr_error')

    return pictur

#### defined in cspyce0_part2.i

def ckfrot_error(inst, et):
    (rotate, ref, found) = cspyce0.ckfrot(inst, et)
    if not found:
        chkin('ckfrot_error')
        setmsg('Orientation data not found for instrument {} at time {}'.format(inst, et))
        sigerr('SPICE(CKINSUFFDATA)')
        chkout('ckfrot_error')

    return [rotate, ref]

def ckfxfm_error(inst, et):
    (rotate, ref, found) = cspyce0.ckfxfm(inst, et)
    if not found:
        chkin('ckfxfm_error')
        setmsg('Orientation data not found for instrument {} at time {}'.format(inst, et))
        sigerr('SPICE(CKINSUFFDATA)')
        chkout('ckfxfm_error')

    return [rotate, ref]

def dafgsr_error(handle, recno, begin, end):
    (data, found) = cspyce0.dafgsr(handle, recno, begin, end)
    if not found:
        chkin('dafgsr_error')
        setmsg('DAF summary content not available for handle {}, '
               'record {}, words {}-{}'.format(handle, recno, begin, end))
        sigerr('SPICE(DAFFRNOTFOUND)')
        chkout('dafgsr_error')

    return data

def dlabbs_error(handle, recno, begin, end):
    (dladsc, found) = cspyce0.dafgsr(handle, recno, begin, end)
    if not found:
        chkin('dlabbs_error')
        setmsg('DLA segment not found for handle {}'.format(handle))
        sigerr('SPICE(DASFILEREADFAILED)')
        chkout('dlabbs_error')

    return dladsc

def dlabfs_error(handle):
    (dladsc, found) = cspyce0.dlabfs(handle)
    if not found:
        chkin('dlabfs_error')
        setmsg('DLA segment not found for handle {}'.format(handle))
        sigerr('SPICE(DASFILEREADFAILED)')
        chkout('dlabfs_error')

    return dladsc

def dlafns_error(handle, dladsc):
    (nxtdsc, found) = cspyce0.dlafns(handle, dladsc)
    if not found:
        chkin('dlafns_error')
        setmsg('DLA segment not found for handle {}'.format(handle))
        sigerr('SPICE(DASFILEREADFAILED)')
        chkout('dlafns_error')

    return nxtdsc

def dlafps_error(handle, dladsc):
    (prvdsc, found) = cspyce0.dlafps(handle, dladsc)
    if not found:
        chkin('dlafps_error')
        setmsg('DLA segment not found for handle {}'.format(handle))
        sigerr('SPICE(DASFILEREADFAILED)')
        chkout('dlafps_error')

    return prvdsc

def dskx02_error(handle, dladsc, vertex, raydir):
    (plid, xpt, found) = cspyce0.dskx02(handle, dladsc, vertex, raydir)
    if not found:
        chkin('dskx02_error')
        setmsg('Intercept plate ID not found')
        sigerr('SPICE(NOINTERCEPT)')
        chkout('dskx02_error')

    return [plid, xpt]

def dskxsi_error(pri, target, nsurf, srflst, et, fixref, vertex, raydir):
    (xpt, handle, dladsc, dskdsc, dc, ic, found) = cspyce0.dskxsi(pri, target, nsurf,
                                                                  srflst, et, fixref,
                                                                  vertex, raydir)
    if not found:
        chkin('dskxsi_error')
        setmsg('Intercept plate ID not found')
        sigerr('SPICE(NOINTERCEPT)')
        chkout('dskxsi_error')

    return [xpt, handle, dladsc, dskdsc, dc, ic]

def ekfind_error(query):
    (nmrows, error, errmsg) = cspyce0.ekfind(query)
    if error:
        chkin('ekfind_error')
        setmsg(errmsg)
        sigerr('SPICE(INVALIDVALUE)')
        chkout('ekfind_error')

    return nmrows

def ekgc_error(selidx, row, elment):
    (cdata, null, found) = cspyce0.ekgc(selidx, row, elment)
    if not found:
        chkin('ekgc_error')
        setmsg('EK item not found: column index {}, '
               'row {}, element {}'.format(selidx, row, elment))
        sigerr('SPICE(INDEXOUTOFRANGE)')
        chkout('ekgc_error')

    return [cdata, null]

def ekgd_error(selidx, row, elment):
    (ddata, null, found) = cspyce0.ekgd(selidx, row, elment)
    if not found:
        chkin('ekgd_error')
        setmsg('EK item not found: column index {}, '
               'row {}, element {}'.format(selidx, row, elment))
        sigerr('SPICE(INDEXOUTOFRANGE)')
        chkout('ekgd_error')

    return [ddata, null]

def ekgi_error(selidx, row, elment):
    (idata, null, found) = cspyce0.ekgi(selidx, row, elment)
    if not found:
        chkin('ekgi_error')
        setmsg('EK item not found: column index {}, '
               'row {}, element {}'.format(selidx, row, elment))
        sigerr('SPICE(INDEXOUTOFRANGE)')
        chkout('ekgi_error')

    return [idata, null]

def ekpsel_error(query, tabs, n4, cols, n5):
    (xbegs, xends, xtypes, xclass, tabs, cols, error, errmsg) = cspyce0.ekpsel(query,
                                                                               tabs, n4,
                                                                               cols, n5)
    if error:
        chkin('ekpsel_error')
        setmsg(errmsg)
        sigerr('SPICE(INVALIDVALUE)')
        chkout('ekpsel_error')

    return [xbegs, xends, xtypes, xclass, tabs, cols]

def hx2dp_error(string):
    (number, error, errmsg) = cspyce0.hx2dp(string)
    if error:
        chkin('hx2dp_error')
        setmsg(errmsg)
        sigerr('SPICE(INVALIDVALUE)')
        chkout('hx2dp_error')

    return number

def kdata_error(which, kind):
    (file, filtype, srcfil, handle, found) = cspyce0.kdata(which, kind)
    if not found:
        chkin('kdata_error')
        count = cspyce0.ktotal(kind)
        if which >= count:
            setmsg('index out of range: {}/{}, {}'.format(which, count, kind))
            sigerr('SPICE(INVALIDINDEX)')
        else:
            setmsg('kernel not found: {}, {}'.format(which, kind))
            sigerr('SPICE(FILENOTFOUND)')
        chkout('kdata_error')

    return [file, filtype, srcfil, handle]

def kinfo_error(file):
    (filtyp, srcfil, handle, found) = cspyce0.kinfo(file)
    if not found:
        chkin('kinfo_error')
        setmsg('kernel file not found: ' + file)
        sigerr('SPICE(FILENOTFOUND)')
        chkout('kinfo_error')

    return [filtyp, srcfil, handle]

def spksfs_error(body, et, idlen, ident):
    (handle, descr, ident, found) = cspyce0.spksfs(body, et, idlen, ident)
    if not found:
        chkin('spksfs_error')
        setmsg('SPK segment not found for body "{}", time {}'.format(body, et))
        sigerr('SPICE(SPKINSUFFDATA)')
        chkout('spksfs_error')

    return [handle, descr, ident]

def szpool_error(name):
    (n, found) = cspyce0.szpool(name)
    if not found:
        chkin('szpool_error')
        setmsg('kernel pool size limit "{}" not found'.format(name))
        sigerr('SPICE(KERNELVARNOTFOUND)')
        chkout('szpool_error')

    return n

def tkfram_error(frcode):
    (rot, frame, found) = cspyce0.tkfram(frcode)
    if not found:
        chkin('tkfram_error')
        setmsg('rotation matrix for frame {} not found'.format(frcode))
        sigerr('SPICE(FRAMEIDNOTFOUND)')
        chkout('tkfram_error')

    return [rot, frame]

#### These functions are both vectorized and have _error versions

def ckgp_vector_error(inst, sclkdp, tol, ref):
    (cmat, clkout, found) = cspyce0.ckgp_vector(inst, sclkdp, tol, ref)
    if not np.all(found):
        name = cspyce0.frmnam(inst)
        if name:
            namestr = ' (' + name + ')'
        else:
            namestr = ''

        sclkdp_min = np.min(sclkdp)
        sclkdp_max = np.max(sclkdp)
        if sclkdp_min == sclkdp_max:
            clockstr = str(sclkdp)
        else:
            clockstr = '%s to %s' % (sclkdp_min, sclkdp_max)

        chkin('ckgp_vector_error')
        setmsg('insufficient C kernel data to evaluate ' +
               'instrument/spacecraft %s%s ' % (inst, namestr) +
               'at spacecraft clock times %s ' % clockstr +
               'with tolerance %s' % np.min(tol))
        sigerr('SPICE(CKINSUFFDATA)')
        chkout('ckgp_vector_error')

    return [cmat, clkout]

def ckgpav_vector_error(inst, sclkdp, tol, ref):
    (cmat, av, clkout, found) = cspyce0.ckgpav_vector(inst, sclkdp, tol, ref)
    if not np.all(found):
        name = cspyce0.frmnam(inst)
        if name:
            namestr = ' (' + name + ')'
        else:
            namestr = ''

        sclkdp_min = np.min(sclkdp)
        sclkdp_max = np.max(sclkdp)
        if sclkdp_min == sclkdp_max:
            clockstr = str(sclkdp)
        else:
            clockstr = '%s to %s' % (sclkdp_min, sclkdp_max)

        chkin('ckgpav_vector_error')
        setmsg('insufficient C kernel data to evaluate ' +
               'instrument/spacecraft %s%s ' % (inst, namestr) +
               'at spacecraft clock times %s ' % clockstr +
               'with tolerance %s' % np.min(tol))
        sigerr('SPICE(CKINSUFFDATA)')
        chkout('ckgpav_vector_error')

    return [cmat, av, clkout]

def invert_error(m1):
    inverse = cspyce0.invert(m1)
    if np.all(inverse == 0.):
        chkin('invert_error')
        setmsg('singular matrix encountered; inverse failed')
        sigerr('SPICE(SINGULARMATRIX)')
        chkout('invert_error')

    return inverse

def invert_vector_error(m1):
    inverses = cspyce0.invert_vector(m1)
    tests = np.all(inverses == 0., (-2,-1))
    if np.any(tests):
        chkin('invert_error')
        setmsg('singular matrix encountered; inverse failed')
        sigerr('SPICE(SINGULARMATRIX)')
        chkout('invert_error')

    return inverses

################################################################################
# This is the one function that takes an array of strings as input. This fix
# allows it to work in a sensible way if a single input string is provided.
################################################################################

def pcpool(name, cvals):
    if isinstance(cvals, str):
        cspyce0.pcpool(name, [cvals])
    else:
        cspyce0.pcpool(name, cvals)

################################################################################
# These wrappers on the comment readers dafec an dasec ensure that the entire
# comment field is read. They eliminate the "done" return value.
################################################################################

def dafec(handle):
    records = []
    while True:
        (buffer, done) = cspyce0.dafec(handle)
        records += buffer
        if done:
            return records

# Update the call signature for help
del CSPYCE_RETURNS['dafec'][1]
del CSPYCE_RETNAMES['dafec'][1]
del CSPYCE_DEFINITIONS['dafec']['done']

def dasec(handle):
    records = []
    while True:
        (buffer, done) = cspyce0.dasec(handle)
        records += buffer
        if done:
            return records

# Update the call signature for help
del CSPYCE_RETURNS['dasec'][1]
del CSPYCE_RETNAMES['dasec'][1]
del CSPYCE_DEFINITIONS['dasec']['done']

################################################################################
# For functions that return only a list of strings, don't embed the results in
# an additional layer [].
################################################################################

def lparse(list_, delim):
    result = cspyce0.lparse(list_, delim)
    if len(result) == 1 and isinstance(result[0], list):
        return result[0]

def lparsm(list_, delims):
    result = cspyce0.lparsm(list_, delims)
    if len(result) == 1 and isinstance(result[0], list):
        return result[0]

################################################################################
# When a function returns a value and a "found" flag, and the flag is False,
# the unused return values contain random values. We change these to something
# sensible.
################################################################################

def bodn2c(name):
    (code, found) = cspyce0.bodn2c(name)
    return [code if found else 0, found]

def bods2c(name):
    (code, found) = cspyce0.bods2c(name)
    return [code if found else 0, found]

def ccifrm(frclss, clssid):
    (frcode, frname, center, found) = cspyce0.ccifrm(frclss, clssid)
    return [frcode if found else 0, frname, center if found else 0, found]

def cnmfrm(cname):
    (frcode, frname, found) = cspyce0.cnmfrm(cname)
    return [frcode if found else 0, frname if found else '', found]

def frinfo(frcode):
    (cent, frclss, clssid, found) = cspyce0.frinfo(frcode)
    return [cent if found else 0,
            frclss if found else 0,
            clssid if found else 0, found]

def srfs2c(srfstr, bodstr):
    (code, found) = cspyce0.srfs2c(srfstr, bodstr)
    return [code if found else 0, found]

def tparse(string):
    (sp2000, msg) = cspyce0.tparse(string)
    return [0. if msg else sp2000, msg]

def ckfrot(inst, et):
    (rotate, ref, found) = cspyce0.ckfrot(inst, et)
    return [rotate if found else np.diag(3*[1.]), ref if found else 0, found]

def ckfxfm(inst, et):
    (rotate, ref, found) = cspyce0.ckfrot(inst, et)
    return [rotate if found else np.diag(6*[1.]), ref if found else 0, found]

def hx2dp(string):
    (number, error, errmsg) = cspyce0.hx2dp(string)
    return [0. if error else number, error, errmsg]

################################################################################
# Handle "GET"/"SET" inputs to timdef().
################################################################################

def timdef(action='', item='', value=''):

    action = action.upper().strip()
    if action not in ('GET', 'SET'):
        if item == '':
            item = action
            action = 'GET'
        else:
            value = item
            item = action
            action = 'SET'

    return cspyce0.timdef(action, item, value)

################################################################################
# Prepare for the possible use of aliases
################################################################################

FRAME_CODE_OVERRIDES = {}
FRAME_NAME_OVERRIDES = {}

def _name_as_key(name):
    """Uppercase, stripped, no repeated interior whitespace."""

    name = name.upper().strip()
    while '  ' in name:
        name = name.replace('  ', ' ')

    return name

def frmnam(frcode):
    return FRAME_NAME_OVERRIDES.get(frcode, cspyce0.frmnam(frcode))

def frmnam_error(frcode):
    return FRAME_NAME_OVERRIDES.get(frcode, frmnam1_error(frcode))

def namfrm(frname):
    return FRAME_CODE_OVERRIDES.get(_name_as_key(frname), cspyce0.namfrm(frname))

def namfrm_error(frname):
    return FRAME_CODE_OVERRIDES.get(_name_as_key(frname), namfrm1_error(frname))

################################################################################
# Assign support information about each function:
#   ABSTRACT    = a brief description of what the function does, formatted
#                 and wrapped at 72-column width. It should start and end with
#                 a newline.
#   SIGNATURE   = a list indicating the type of each argument
#   ARGNAMES    = a list of names of the input arguments
#   RETURNS     = a list indicating the type of each return value
#   RETNAMES    = a list of names of the return values
#   URL         = a URL for more information
#   PS          = an optional, final note.
#   DEFINITIONS = a dictionary describing each input and return values.
#   NOTES       = a list of one or more note strings, each formatted and
#                 wrapped at 72-column width. It should start and end with a
#                 newline.
#
# Note that the docstring and defaults will be constructed from this
# information.
################################################################################

def assign_docstring(func, note=""):
    """Assign a docstring to a cspyce function.

    An optional additional note string can provide version information. It must
    be pre-formatted, wrapped at 72 columns, and begin and end with a newline.
    """

    doclist = [func.ABSTRACT, '\n']

    if func.URL:
        doclist += [func.URL, '\n']

    if note:
        # We need to copy the note list so as not to affect other versions using
        # the same note list.
        func.NOTES = list(func.NOTES) + [note]

    doclist += func.NOTES

    lname = 0
    names = []
    for name in func.ARGNAMES + func.RETNAMES:
        lname = max(lname, len(name))
        names.append(name)

    ltype = 0
    types = []
    for arg_type in func.SIGNATURE + func.RETURNS:
        arg_type = (arg_type.replace('time', 'float')
                            .replace('rotmat', 'float')
                            .replace('body_code', 'int')
                            .replace('body_name', 'string')
                            .replace('frame_code', 'int')
                            .replace('frame_name', 'string'))
        ltype = max(ltype, len(arg_type))
        types.append(arg_type)

    indent = 2 + ltype + 1 + lname + 3
    ldefs = 72 - indent
    tabstr = indent * ' '

    inputs = len(func.ARGNAMES)
    doclist += ['\nInputs:']
    if inputs == 0:
        doclist += [' none\n']
    else:
        doclist += ['\n']

    for (name, arg_type) in zip(names[:inputs], types[:inputs]):
        desc = textwrap.wrap(func.DEFINITIONS[name][1], ldefs)
        doclist += ['  ', arg_type, (ltype - len(arg_type))*' ', ' ']
        doclist += [name, (lname - len(name))*' ', ' = ']
        doclist += [desc[0], '\n']
        for k in range(1, len(desc)):
            doclist += [tabstr, desc[k], '\n']

    doclist += ['\nReturns:']
    if len(func.RETNAMES) == 0:
        doclist += [' none\n']
    else:
        doclist += ['\n']

    for (name, arg_type) in zip(names[inputs:], types[inputs:]):
        desc = textwrap.wrap(func.DEFINITIONS[name][1], ldefs)
        doclist += ['  ', arg_type, (ltype - len(arg_type))*' ', ' ']
        doclist += [name, (lname - len(name))*' ', ' = ']
        doclist += [desc[0], '\n']
        for k in range(1, len(desc)):
            doclist += [tabstr, desc[k], '\n']

    if func.PS:
        ps = textwrap.wrap('Note: ' + func.PS)
        doclist += ['\n', '\n'.join(ps)]

    doclist += ['\n']

    func.__doc__ = ''.join(doclist)

# Non-vector functions
for name in CSPYCE_SIGNATURES:
    func = globals()[name]
    func.ABSTRACT  = CSPYCE_ABSTRACT[name]
    func.SIGNATURE = CSPYCE_SIGNATURES[name]
    func.ARGNAMES  = CSPYCE_ARGNAMES[name]
    func.RETURNS   = CSPYCE_RETURNS[name]
    func.RETNAMES  = CSPYCE_RETNAMES[name]
    func.PS        = CSPYCE_PS.get(name, '')
    func.URL       = CSPYCE_URL.get(name, '')
    func.NOTES     = []
    func.DEFINITIONS = CSPYCE_DEFINITIONS[name]

    if name in CSPYCE_DEFAULTS:
        if PYTHON2:
            func.func_defaults = tuple(CSPYCE_DEFAULTS[name])
        else:
            func.__defaults__ = tuple(CSPYCE_DEFAULTS[name])

    assign_docstring(func)

# This is a set of every unique cspyce function's basename (before any suffix)
CSPYCE_BASENAMES = {n for n in CSPYCE_ABSTRACT.keys()
                    if not n.endswith('_error')}

################################################################################
# Assign docstrings, signatures, etc. to the _vector functions
################################################################################

# arg_type -> (arg_type for input, arg_type for output)
VECTORIZED_ARGS = {
    'time'       : ('time[_]'      , 'time[_]'      ),
    'float'      : ('float[_]'     , 'float[_]'     ),
    'float[2]'   : ('float[_,2]'   , 'float[_,2]'   ),
    'float[3]'   : ('float[_,3]'   , 'float[_,3]'   ),
    'float[4]'   : ('float[_,4]'   , 'float[_,4]'   ),
    'float[6]'   : ('float[_,6]'   , 'float[_,6]'   ),
    'float[8]'   : ('float[_,8]'   , 'float[_,8]'   ),
    'float[9]'   : ('float[_,9]'   , 'float[_,9]'   ),
    'float[2,2]' : ('float[_,2,2]' , 'float[_,2,2]' ),
    'float[3,3]' : ('float[_,3,3]' , 'float[_,3,3]' ),
    'float[6,6]' : ('float[_,6,6]' , 'float[_,6,6]' ),
    'float[*]'   : ('float[_,*]'   , 'float[_,*]'   ),
    'float[*,*]' : ('float[_,*,*]' , 'float[_,*,*]' ),
    'rotmat[3,3]': ('rotmat[_,3,3]', 'rotmat[_,3,3]'),
    'rotmat[6,6]': ('rotmat[_,6,6]', 'rotmat[_,6,6]'),
    'int'        : ('int'          , 'int[_]'       ),
    'bool'       : ('bool'         , 'bool[_]'      ),
    'body_code'  : ('body_code'    , 'body_code[_]' ),  # not used (yet)
    'frame_code' : ('frame_code'   , 'frame_code[_]'),  # not used (yet)
    'body_name'  : ('body_name'    , 'body_name[_]' ),  # not used (yet)
    'frame_name' : ('frame_name'   , 'frame_name[_]'),  # not used (yet)
}

def _vectorize_signature(signature):
    """Convert a scalar signature to a vectorized signature."""
    return _vectorize_arglist(signature, 0)

def _vectorize_return(returns):
    """Convert a scalar return list to a vectorized return list."""
    return _vectorize_arglist(returns, 1)

def _vectorize_arglist(arglist, k):
    vectorized = []
    for arg in arglist:
        if arg in VECTORIZED_ARGS:
            vectorized.append(VECTORIZED_ARGS[arg][k])
        else:
            vectorized.append(arg)

    return vectorized

VECTOR_NOTE = """
In this vectorized version, any or all of the floating-point inputs can
have an extra leading dimension. The function will loop over this axis
and return arrays of the results. If no inputs have an extra dimension,
it returns results identical to the un-vectorized version.
"""

for basename in CSPYCE_BASENAMES:
    for suffix in ('', '_error'):
        vname = basename + '_vector' + suffix
        if vname not in globals():
            continue

        name = basename + suffix
        func = globals()[name]
        vfunc = globals()[vname]

        vfunc.ABSTRACT = func.ABSTRACT
        vfunc.SIGNATURE = _vectorize_signature(func.SIGNATURE)
        vfunc.ARGNAMES = func.ARGNAMES
        vfunc.RETURNS = _vectorize_signature(func.RETURNS)
        vfunc.RETNAMES = func.RETNAMES
        vfunc.PS        = func.PS
        vfunc.URL       = func.URL
        vfunc.DEFINITIONS = func.DEFINITIONS
        vfunc.NOTES     = func.NOTES

        assign_docstring(vfunc, VECTOR_NOTE)

################################################################################
# Register _flag, _error, _scalar, and _vector versions of every function.
################################################################################

# The only options are:
#   func == func_error == func_vector
#   func == func_error != func_vector
#   func == func_vector != func_error
#   func != func_vector != func_error != func_vector_error

for name in CSPYCE_BASENAMES:
    ename  = name + '_error'
    vname  = name + '_vector'
    vename = name + '_vector_error'

    # Get up to four versions of the function
    func   = globals()[name]
    efunc  = globals().get(ename, func)
    vfunc  = globals().get(vname, func)

    if vename in globals():
        vefunc = globals()[vename]
    elif vfunc is func:
        vefunc = efunc
    elif efunc is func:
        vefunc = vfunc
    else:
        vefunc = func

    # Define them globally. They may already be defined, in which case this is
    # a harmless operation
# DISABLED! It creates too many symbols for the global dictionary
#     globals()[ ename] =  efunc
#     globals()[ vname] =  vfunc
#     globals()[vename] = vefunc

    # Define the links between the different versions.
    func.flag   =  func
    func.error  = efunc
    func.vector = vfunc
    func.scalar =  func

    efunc.flag   =   func
    efunc.error  =  efunc
    efunc.vector = vefunc
    efunc.scalar =  efunc

    vfunc.flag   =  vfunc
    vfunc.error  = vefunc
    vfunc.vector =  vfunc
    vfunc.scalar =   func

    vefunc.flag   =  vfunc
    vefunc.error  = vefunc
    vefunc.vector = vefunc
    vefunc.scalar =  efunc

    # Define the alternative names for these functions
# DISABLED! It creates too many symbols for the global dictionary
#     fname  = name + '_flag'
#     sname  = name + '_scalar'
#     sfname = name + '_scalar_flag'
#     sename = name + '_scalar_error'
#     vfname = name + '_vector_flag'
#
#     globals()[ fname] =  func
#     globals()[ sname] =  func
#     globals()[sfname] =  func
#     globals()[sename] = efunc
#     globals()[vfname] = vfunc

################################################################################
# Set defaults at initialization
################################################################################

erract('SET', 'EXCEPTION')

################################################################################



################################################################################
# cspyce/cspyce2.py
#
# This module re-declares every cspyce1 function explicitly, with its list of
# argument names as used by CSPICE. The practical effect is that functions in
# cspyce2 module can be called in a fully Python-like way, the rightmost inputs
# in any order and identified by their names.
#
# Used internally by cspyce; not intended for direct import.
################################################################################

import sys
import cspyce.cspyce1 as cspyce1
import keyword

PYTHON2 = sys.version_info[0] < 3

# This function makes cspyce2 look the same as cspyce1. It ensures that every
# location in the global dictionary and every function's internal link point to
# a new function of the same name.

CSPYCE_FUNCTION_LOOKUP = {}


def relink_all(new_dict, old_dict):

    # Assign a new function to the dictionary at the same location as every
    # cspyce function found in the old dictionary

    dict_names = {}     # maps each function name to its dictionary locations
    old_funcs = {}      # maps each function name to its old function
    for (dict_name, old_func) in old_dict.items():
        if not callable(old_func):
            continue
        if 'SIGNATURE' not in old_func.__dict__:
            continue

        func_name = old_func.__name__
        old_funcs[func_name] = old_func

        if func_name not in dict_names:
            dict_names[func_name] = []

        dict_names[func_name].append(dict_name)

    for (name, keys) in dict_names.items():
        func = new_dict[name]
        for key in keys:
            new_dict[key] = func
            CSPYCE_FUNCTION_LOOKUP[func.__name__] = func

    # Make sure each cspyce function has the same properties and attributes as
    # the one in the old dictionary

    for (name, old_func) in old_funcs.items():
        func = new_dict[name]

        # Copy function properties
        func.__doc__ = old_func.__doc__

        if PYTHON2:
            func.func_defaults = old_func.func_defaults
        else:
            func.__defaults__ = old_func.__defaults__

        # Copy attributes
        for (key, value) in old_func.__dict__.items():
            if not callable(value):
                func.__dict__[key] = value
            else:
                # If it's a function, locate a new one with the same name
                func.__dict__[key] = new_dict[value.__name__]


def populate_cspyce2():
    snippets = []
    for name, func in cspyce1.__dict__.items():
        if callable(func) and hasattr(func, 'ARGNAMES'):
            argnames = [(x + "_" if keyword.iskeyword(x) else x) for x in func.ARGNAMES]
            code = "def {name}({arglist}): return cspyce1.{name}({arglist})".format(
                       name=name, arglist=", ".join(argnames))
            snippets.append(code)
    code = "\n\n".join(snippets)
    exec(code, globals(), globals())


populate_cspyce2()
relink_all(globals(), cspyce1.__dict__)

################################################################################



import unittest

import numpy as  np
import cspyce.typemap_samples as ts


def flatten(array):
    return tuple(tuple(array.ravel()))

# noinspection PyTypeChecker
class test_array1_1(unittest.TestCase):
    # %apply (int IN_ARRAY1[ANY]) {int arg[3]}
    # cs.in_array1_1 just returns whatever 3 integers it was passed as a numpy array
    def test_basic_test_tuple(self):
        self.assertEqual((1, 2, 3), ts.in_array1_1((1, 2, 3)))

    def test_basic_test_array(self):
        array = np.arange(10, 13, dtype="int32")
        self.assertEqual(tuple(array), ts.in_array1_1(array))

    def test_non_contiguous_array(self):
        array = np.arange(9, dtype='int32').reshape((3, 3))
        self.assertEqual((0, 3, 6), ts.in_array1_1(array[:, 0]))

    def test_requires_three_elements(self):
        with self.assertRaises(ValueError):
            ts.in_array1_1((1, 2, 3, 4))

    def test_requires_integer_array(self):
        with self.assertRaises(ValueError):
            ts.in_array1_1(np.arange(3.0))

    def test_requires_one_dimensional_array(self):
        array = np.zeros((3, 3), dtype="int32")
        with self.assertRaises(ValueError):
            ts.in_array1_1(array)

    def test_requires_non_null(self):
        with self.assertRaises(TypeError):
            ts.in_array1_1(None)


class test_array1_2(unittest.TestCase):
    # %apply (int* IN_ARRAY1, int DIM1) {(int* arg, int dim)};
    # This function returns them as a list.
    def test_basic_test_tuple(self):
        self.assertEqual((1, 2, 3, 4, 5), ts.in_array1_2([1, 2, 3, 4, 5]))

    def test_basic_test_array(self):
        arg = np.arange(1, 10, dtype='int32')
        self.assertEqual(flatten(arg), ts.in_array1_2(arg))

    def test_okay_to_pass_empty_list(self):
        self.assertEqual((), ts.in_array1_2(()))

    def test_non_contiguous_array(self):
        array = np.arange(12, dtype='int32').reshape((4, 3))[:, 1]
        self.assertEqual(flatten(array), ts.in_array1_2(array))

    def test_requires_integer_array(self):
        with self.assertRaises(ValueError):
            ts.in_array1_2(np.arange(10.0))

    def test_requires_one_dimensional_array(self):
        array = np.zeros((3, 3), dtype="int32")
        with self.assertRaises(ValueError):
            ts.in_array1_2(array)

    def test_requires_one_dimensional_int_array(self):
        array = np.zeros((3, 3), dtype="double")
        with self.assertRaises(ValueError):
            ts.in_array1_2(array)

    def test_requires_non_null(self):
        with self.assertRaises(TypeError):
            ts.in_array1_2(None)

class test_array1_3(unittest.TestCase):
    # %apply (int* IN_ARRAY1, int DIM1) {(int* arg)};
    # This function is exactly like test_array_1_2, except the length is passed separately
    def test_basic_test_tuple(self):
        self.assertEqual((1, 2, 3, 4, 5), ts.in_array1_3([1, 2, 3, 4, 5], 5))

    def test_basic_test_array(self):
        arg = np.arange(1, 10, dtype='int32')
        self.assertEqual(flatten(arg), ts.in_array1_3(arg, len(arg)))

    def test_okay_to_pass_empty_list(self):
        self.assertEqual((), ts.in_array1_3((), 0))

    def test_non_contiguous_array(self):
        array = np.arange(12, dtype='int32').reshape((4, 3))[:, 1]
        self.assertEqual(flatten(array), ts.in_array1_3(array, len(array)))

    def test_requires_integer_array(self):
        with self.assertRaises(ValueError):
            ts.in_array1_3(np.arange(10.0), 1)

    def test_requires_one_dimensional_array(self):
        array = np.zeros((3, 3), dtype="int32")
        with self.assertRaises(ValueError):
            ts.in_array1_3(array, 9)

    def test_requires_one_dimensional_int_array(self):
        array = np.zeros((3, 3), dtype="double")
        with self.assertRaises(ValueError):
            ts.in_array1_3(array, 9)

    def test_requires_non_null(self):
        with self.assertRaises(TypeError):
            ts.in_array1_3(None, 0)


class test_array1_01_1(unittest.TestCase):
    # %apply (int *IN_ARRAY01, int DIM1) {(int *arg, int dim)};
    # cs.in_array01_1 received either an int scalar or sequence of integer, and
    SMALL_INT_ARRAY = np.array((4, 5, 6), dtype="int32")
    SMALL_FLOAT_ARRAY = np.array((4.0, 5.0, 6.0), dtype="double")

    def f(self, x):
        return ts.in_array01_1(x)

    def test_basic_test_scalar(self):
        self.assertEqual(1, ts.in_array01_1(1))

    def test_basic_test_tuple(self):
        self.assertEqual((1, 2, 3), ts.in_array01_1([1, 2, 3]))

    def test_basic_test_int_array(self):
        arg = np.arange(1, 10, dtype='int32')
        self.assertEqual(flatten(arg), ts.in_array01_1(arg))

    def test_non_contiguous_array(self):
        array = np.arange(12, dtype='int32').reshape((4, 3))[:, 0]
        self.assertEqual(flatten(array), ts.in_array01_1(array))

    def test_requires_integer_array(self):
        with self.assertRaises(ValueError):
            ts.in_array01_1(np.arange(20.))

    def test_requires_one_dimensional_array(self):
        array = np.zeros((3, 3), dtype="int32")
        with self.assertRaises(ValueError):
            ts.in_array01_1(array)

    def test_requires_one_dimensional_int_array(self):
        array = np.zeros((3, 3), dtype="double")
        with self.assertRaises(ValueError):
            ts.in_array01_1(array)

    def test_requires_non_null(self):
        with self.assertRaises(TypeError):
            ts.in_array01_1(None)


class test_array2_1(unittest.TestCase):
    # %apply (int IN_ARRAY2[ANY][ANY]) {int arg[3][5]};
    # This function specifically requires a 3x5 int array.
    # It returns the first element, and the dimensions as a tuple.
    def test_basic_run(self):
        array = np.arange(1000, 1015, dtype='int32').reshape(3, 5)
        self.assertEqual((flatten(array), 3, 5), ts.in_array2_1(array))

    def test_non_contiguous_array(self):
        array = np.arange(150, dtype='int32').reshape((3, 5, 10))[..., 2]
        self.assertEqual((flatten(array), 3, 5), ts.in_array2_1(array))

    def test_no_other_size(self):
        array = np.array(range(100, 115), dtype='int32').reshape(5, 3)
        with self.assertRaises(ValueError):
            ts.in_array2_1(array)

    def test_no_other_data_type(self):
        array = np.array(range(100, 115), dtype='int64').reshape(3, 5)
        with self.assertRaises(ValueError):
            ts.in_array2_1(array)

    def test_no_other_dimension(self):
        array = np.array(range(100, 115), dtype='int64').reshape(3, 5, 1)
        with self.assertRaises(ValueError):
            ts.in_array2_1(array)

    def test_requires_non_null(self):
        with self.assertRaises(TypeError):
            ts.in_array2_1(None)


class test_array2_2(unittest.TestCase):
    # %apply (int *IN_ARRAY2, int DIM1, int DIM2) {(int *arg, int dim1, int dim2)};
    # This function takes any sized integer array.
    # It returns the elements of the array as a tuple, and the dimensions
    def test_basic_run(self):
        array = np.arange(100, 200, dtype='int32').reshape(5, 20)
        self.assertEqual((flatten(array), 5, 20), ts.in_array2_2(array))
        self.assertEqual((flatten(array[1:]), 4, 20), ts.in_array2_2(array[1:]))

    def test_non_contiguous_array(self):
        array = np.arange(150, dtype='int32').reshape((3, 5, 10))[..., 2]
        self.assertEqual((flatten(array), 3, 5), ts.in_array2_2(array))

    def test_no_other_data_type(self):
        array = np.arange(100., 200.).reshape(5, 20)
        with self.assertRaises(ValueError):
            ts.in_array2_2(array)

    def test_no_bigger_dimension(self):
        array = np.arange(100, 200, dtype='int32').reshape(5, 20, 1)
        with self.assertRaises(ValueError):
            ts.in_array2_2(array)

    def test_no_smaller_dimension(self):
        array = np.arange(100, 200, dtype='int32')
        with self.assertRaises(ValueError):
            ts.in_array2_2(array)

    def test_requires_non_null(self):
        with self.assertRaises(TypeError):
            ts.in_array2_2(None)


class test_array2_3(unittest.TestCase):
    # %apply (int IN_ARRAY2[][ANY], int DIM1) {(int arg[][5], int dim1)};
    # This function takes any 2-dimensional array whose second dimension is 5.
    # It returns the elements, and the dimensions.
    def test_basic_run(self):
        array = np.arange(100, 150, dtype='int32').reshape(10, 5)
        self.assertEqual((flatten(array), 10, 5), ts.in_array2_3(array))
        self.assertEqual((flatten(array[1:]), 9, 5), ts.in_array2_3(array[1:]))

    def test_non_contiguous_array(self):
        array = np.arange(150, dtype='int32').reshape((3, 5, 10))[..., 2]
        self.assertEqual((flatten(array), 3, 5), ts.in_array2_3(array))

    def test_no_other_width(self):
        array = np.arange(100, 150, dtype='int32').reshape(5, 10)
        with self.assertRaises(ValueError):
            ts.in_array2_3(array)

    def test_no_other_data_type(self):
        array = np.arange(100, 150, dtype='float').reshape(10, 5)
        with self.assertRaises(ValueError):
            ts.in_array2_3(array)

    def test_no_bigger_dimension(self):
        array = np.arange(100, 150, dtype='int32').reshape(1, 10, 5)
        with self.assertRaises(ValueError):
            ts.in_array2_3(array)

    def test_no_smaller_dimension(self):
        array = np.arange(100, 150, dtype='int32')
        with self.assertRaises(ValueError):
            ts.in_array2_3(array)

    def test_requires_non_null(self):
        with self.assertRaises(TypeError):
            ts.in_array2_3(None)

class test_array2_3(unittest.TestCase):
    # %apply (int IN_ARRAY2[][ANY], int DIM1) {(int arg[][5], int dim1)};
    # This function takes any 2-dimensional array whose second dimension is 5.
    # It returns the elements, and the dimensions.
    def test_basic_run(self):
        array = np.arange(100, 150, dtype='int32').reshape(10, 5)
        self.assertEqual((flatten(array), 10, 5), ts.in_array2_3(array))
        self.assertEqual((flatten(array[1:]), 9, 5), ts.in_array2_3(array[1:]))

    def test_non_contiguous_array(self):
        array = np.arange(150, dtype='int32').reshape((3, 5, 10))[..., 2]
        self.assertEqual((flatten(array), 3, 5), ts.in_array2_3(array))

    def test_no_other_width(self):
        array = np.arange(100, 150, dtype='int32').reshape(5, 10)
        with self.assertRaises(ValueError):
            ts.in_array2_3(array)

    def test_no_other_data_type(self):
        array = np.arange(100, 150, dtype='float').reshape(10, 5)
        with self.assertRaises(ValueError):
            ts.in_array2_3(array)

    def test_no_bigger_dimension(self):
        array = np.arange(100, 150, dtype='int32').reshape(1, 10, 5)
        with self.assertRaises(ValueError):
            ts.in_array2_3(array)

    def test_no_smaller_dimension(self):
        array = np.arange(100, 150, dtype='int32')
        with self.assertRaises(ValueError):
            ts.in_array2_3(array)

    def test_requires_non_null(self):
        with self.assertRaises(TypeError):
            ts.in_array2_3(None)


class test_array12(unittest.TestCase):
    # %apply (int *IN_ARRAY12, int DIM1, int DIM2)  {(int *arg, int dim1, int dim2)};
    # This function can take one or two dimensional arrays.  If only one dimension,
    # then dim1 == 0.  As usual, this returns the array elements, and the dimension
    def test_run_1d(self):
        array = np.arange(100, 150, dtype='int32')
        self.assertEqual((flatten(array), 0, array.size), ts.in_array12(array))

    def test_run_2d(self):
        array = np.arange(100, 150, dtype='int32').reshape(10, 5)
        expected_result = (flatten(array),) + tuple(array.shape)
        self.assertEqual(expected_result, ts.in_array12(array))

    def test_run_on_tuple(self):
        value = (2, 3, 5, 7, 11)
        self.assertEqual((value, 0, len(value)), ts.in_array12(value))

    def test_non_contiguous_array_2d(self):
        array = np.arange(150, dtype='int32').reshape((3, 5, 10))[..., 1]
        expected_result = (flatten(array),) + tuple(array.shape)
        self.assertEqual(expected_result, ts.in_array12(array))

    def test_non_contiguous_array_1d(self):
        array = np.arange(150, dtype='int32').reshape((3, 5, 10))[1, :, 2]
        self.assertEqual((flatten(array), 0, array.size), ts.in_array12(array))

    def test_no_other_data_type(self):
        array = np.arange(100, 150, dtype='double')
        with self.assertRaises(ValueError):
            ts.in_array12(array)

    def test_no_bigger_dimensions(self):
        array = np.arange(100, 150, dtype='int32').reshape(2, 5, 5)
        with self.assertRaises(ValueError):
            ts.in_array12(array)

    def test_requires_non_null(self):
        with self.assertRaises(TypeError):
            ts.in_array12(None)

class test_array23(unittest.TestCase):
    # %apply (int *IN_ARRAY23, int DIM2, int DIM3)  {(int *arg, int dim1, int dim2, int dim3)};
    # This function can take one or two dimensional arrays.  If only one dimension,
    # then dim1 == 0.  As usual, this returns the array elements, and the dimension
    def test_run_2d(self):
        array = np.arange(100, 150, dtype='int32').reshape(10, 5)
        expected_result = (flatten(array), 0) + tuple(array.shape)
        self.assertEqual(expected_result, ts.in_array23(array))

    def test_run_3d(self):
        array = np.arange(100, 150, dtype='int32').reshape(10, 1, 5)
        expected_result = (flatten(array),) + tuple(array.shape)
        self.assertEqual(expected_result, ts.in_array23(array))

    def test_non_contiguous_array_2d(self):
        array = np.arange(150, dtype='int32').reshape((3, 5, 10))[..., 1]
        self.assertEqual((flatten(array), 0, 3, 5), ts.in_array23(array))

    def test_non_contiguous_array_3d(self):
        array = np.empty((3, 4, 5, 8), dtype='int32')[...,2]
        self.assertEqual((flatten(array), 3, 4, 5), ts.in_array23(array))

    def test_no_other_data_type(self):
        array = np.arange(100, 150, dtype='double').reshape(10, 5)
        with self.assertRaises(ValueError):
            ts.in_array23(array)

    def test_no_bigger_dimensions(self):
        array = np.arange(100, 200, dtype='int32').reshape(2, 5, 5, 2)
        with self.assertRaises(ValueError):
            ts.in_array23(array)

    def test_requires_non_null(self):
        with self.assertRaises(TypeError):
            ts.in_array23(None)


class test_out_array1(unittest.TestCase):
    # %apply (int OUT_ARRAY1[ANY]) {(int array[100])};
    # This function fills up a 100-element int array starting at the passed argument
    def test_fixed_size_array(self):
        value = ts.out_array1_1(200)
        self.assertEqual(list(range(200, 300)), list(value))

    # %apply (int OUT_ARRAY1[ANY], int *SIZE1) {(int array[100], int *size)};
    # This function can fill an array up to size 100.  The arguments are start and length
    def test_variable_size_array(self):
        value = ts.out_array1_2(200, 5)
        self.assertEqual(len(value), 5)
        self.assertEqual(value[0], 200)
        self.assertEqual(value[-1], 204)

class test_out_array1_malloced_array(unittest.TestCase):
    # %apply (int **OUT_ARRAY1, int *SIZE1) {(int **arrayP, int *size)};
    # The function mallocs an array of whatever side it needs uses the arguments as
    # start and length
    def test_malloced_array(self):
        value = ts.out_array1_malloc(1000, 5000)
        self.assertEqual(len(value), 5000)
        self.assertEqual(value[0], 1000)
        self.assertEqual(value[-1], 6000 - 1)

    def test_malloced_array_memory_failure(self):
        with self.assertRaises(MemoryError):
            ts.out_array1_malloc(-1, 5000)


class test_out_array01_malloced_array(unittest.TestCase):
    # %apply (double **OUT_ARRAY01, int *SIZE1) {(double **arrayP, int *size)};
    # Again, a malloced array, but if the function indicates size 0, then we want a scalar
    # Again, start and length, but this time we use floats
    def test_return_scalar(self):
        result = ts.out_array01_malloc(5.0, 0)
        self.assertEqual(float, type(result))
        self.assertEqual(5.0, result)

    def test_return_1d(self):
        result = ts.out_array01_malloc(5.0, 2)
        self.assertEqual((5.0, 6.0), flatten(result))

    def test_memory_error(self):
        with self.assertRaises(MemoryError):
            result = ts.out_array01_malloc(-1.0, 10)


class test_out_array2(unittest.TestCase):
    # %apply (int OUT_ARRAY2[ANY][ANY]) {(int array[2][3])};
    # returns a fixed size array filled with the indicated starting number
    def test_fixed_size_array(self):
        # Returns a fixed sized 2x3 array filled with integers starting from 100
        value = ts.out_array2_1(100)
        self.assertEqual((2, 3), value.shape)
        self.assertEqual(value[0,0], 100)

    # %apply (int OUT_ARRAY2[ANY][ANY], int *SIZE1) {(int array[1000][2], int *size)};
    # returns an Nx2 array.  Arguments are start value and length
    def test_variable_size_array(self):
        value = ts.out_array2_2(30, 10)
        self.assertEqual((10, 2), value.shape)
        self.assertEqual(30, value[0, 0])

    # %apply (int **OUT_ARRAY12, int *SIZE1, int *SIZE2) {(int **result, int *size1, int *size2)};
    # Malloced array.  Our function has arguments start value, and dim1, dim2
    # If the start value is < 0, instead, it simulates a malloc failure
    def test_malloced_array(self):
        value = ts.out_array2_3(25, 40, 41)
        self.assertEqual((40, 41), value.shape)
        self.assertEqual(25, value[0, 0])

    def test_memory_error(self):
        with self.assertRaises(MemoryError):
            value = ts.out_array2_3(-1, 40, 41)

    # %apply (int DIM1, int *SIZE1, double OUT_ARRAY2[ANY][ANY]) {(int dim1, int *size1, double result[4][5])};
    def test_2dim_fixed_size_array(self):
        array_length, result = ts.out_array2_4(10, 3)
        self.assertEqual(4, array_length);  # Why does the function need this?
        self.assertEqual((3, 5), result.shape)
        self.assertEqual(flatten(np.arange(10.0, 25.0)), flatten(result))

    # %apply (int *SIZE1, double OUT_ARRAY2[ANY][ANY]) {(int *size1, double result[4][5])};
    def test_2dim_fixed_size_array(self):
        result = ts.out_array2_5(2)
        self.assertEqual((2, 5), result.shape)
        # This generates a boolean array in which the elemnts whose index is a multiple of 3
        # is true.  Just not worth dealing with this

class test_out_array12(unittest.TestCase):
    # %apply (int **OUT_ARRAY12, int *SIZE1, int *SIZE2) {(int **result, int *size1, int *size2)};
    # Same as before, but a dim1=0 indicates to return a 1-dimensional array
    def test_2d_array(self):
        value1 = ts.out_array12_1(25, 40, 41)
        self.assertEqual((40, 41), value1.shape)
        self.assertEqual(25, value1[0, 0])

    def test_generating_1d_array(self):
        value2 = ts.out_array12_1(25, 0, 41)
        self.assertEqual((41,), value2.shape)
        self.assertEqual(25, value2[0])

    def test_memory_error(self):
        with self.assertRaises(MemoryError):
            ts.out_array2_3(-1, 40, 41)


class test_out_array23(unittest.TestCase):
    # %apply (double **OUT_ARRAY23, int *SIZE1, int *SIZE2, int *SIZE3) {(double **result, int *size1, int *size2, int *size3)};
    # Same as before, but 2 or 3 dimensions
    def test_yields_3d(self):
        value1 = ts.out_array23_1(25, 3, 4, 5)
        self.assertEqual((3, 4, 5), value1.shape)
        self.assertEqual(25, value1[0, 0, 0])

    def test_yields_2d(self):
        value2 = ts.out_array23_1(25, 0, 4, 5)
        self.assertEqual((4, 5), value2.shape)
        self.assertEqual(25, value2[0, 0])

    def test_memory_error(self):
        with self.assertRaises(MemoryError):
            ts.out_array23_1(-1, 0, 4, 5)

class test_inout_array_1d(unittest.TestCase):
    # double_in_out_array doubles each element in the array.
    # cs.in_array1_1 just returns whatever 3 integers it was passed as a numpy array
    def test_basic_test_tuple(self):
        self.assertEqual((2, 4, 6), flatten(ts.double_in_out_array((1, 2, 3))))

    def test_basic_test_array(self):
        array = np.arange(10, 13, dtype="int32")
        self.assertEqual((20, 22, 24), flatten(ts.double_in_out_array(array)))
        # verify that array itself hasn't been touched
        self.assertEqual((10, 11, 12), flatten(array))

    def test_non_contiguous_array(self):
        array = np.arange(9, dtype='int32').reshape((3, 3))
        self.assertEqual((0, 6, 12), flatten(ts.double_in_out_array(array[:, 0])))

    def test_requires_integer_array(self):
        with self.assertRaises(ValueError):
            ts.double_in_out_array(np.arange(3.0))

    def test_requires_one_dimensional_array(self):
        array = np.zeros((3, 3), dtype="int32")
        with self.assertRaises(ValueError):
            ts.double_in_out_array(array)

    def test_requires_non_null(self):
        with self.assertRaises(TypeError):
            ts.double_in_out_array(None)



class test_single_strings_input(unittest.TestCase):
    # %apply (char *CONST_STRING) {(const char *string)};
    # our function just returns the length
    def test_const_string(self):
        self.assertEqual(5, ts.const_string_0("abcde"))

    def test_const_string_bad_input(self):
        with self.assertRaises(ValueError):
            ts.const_string_0(23)

    def test_const_string_null_input(self):
        with self.assertRaises(ValueError):
            ts.const_string_0(None)

    # %apply (char IN_STRING) {(char value)};
    # our function accepts a character, and returns it as its integer value
    def test_const_char_0(self):
        self.assertEqual(ord('a'), ts.const_char_0('a'))

    def test_const_char_0_must_be_char(self):
        with self.assertRaises(ValueError):
            ts.const_char_0(3.14)

    def test_const_char_0_cant_be_string(self):
        with self.assertRaises(ValueError):
            ts.const_char_0("abcd")

    def test_const_char_0_cant_be_NULL(self):
        with self.assertRaises(ValueError):
            ts.const_char_0(None)


class test_single_strings_output(unittest.TestCase):

    def test_inout_string(self):
        # %apply (int DIM1, char INOUT_STRING[ANY]) {(int dim, char result[10])};
        # Has a buffer that's a minimum size of 10, but we grow it if necessary for the
        # argument.  We return the size of the buffer
        self.assertEqual("10", ts.inout_string("abcd"))
        self.assertEqual("101", ts.inout_string("a" * 100))

    def test_inout_string_bad_argument(self):
        with self.assertRaises(ValueError):
            ts.inout_string(5.8)

    def test_out_string(self):
        # %apply (char OUT_STRING[ANY]) {(char result[10])};
        # A fixed size buffer.  We return a string representing the value passed
        self.assertEqual("23", ts.out_string(23))


class test_multiple_strings_input(unittest.TestCase):
    # %apply (char *IN_STRINGS, int DIM1, int DIM2) {(const char *strings, int dim1, int dim2)};
    # The string arguments are packed into an N (count) by M (longest length + 1) buffer.
    # This program recreats the original strings and also returns the value of M.
    def test_in_strings(self):
        args = ("abcdefg", "a * 100", "", "xyz" * 35)
        result = ts.in_strings(args)
        self.assertEqual(args, result[:-1])
        self.assertEqual((len(args), 1 + max(len(x) for x in args)), result[-1])

    def test_in_strings_bad_must_strings(self):
        with self.assertRaises(ValueError):
            ts.in_strings(("abc", 27))

    def test_in_strings_bad_must_be_strings_2(self):
        with self.assertRaises(ValueError):
            ts.in_strings(np.array((1, 2, 3)))

    def test_in_strings_cant_be_None(self):
        with self.assertRaises(TypeError):
            ts.in_strings(None)

    def test_in_strings_cant_contain_None(self):
        with self.assertRaises(ValueError):
            ts.in_strings(("abc", None, "DEF"))


class test_multiple_strings_output(unittest.TestCase):
    # %apply (int DIM1, int DIM2, int *NSTRINGS, SpiceChar OUT_STRINGS[ANY][ANY])
    #          {(int dim1, int dim2, int *size, char buffer[50][256])};
    # We are given a 50x256 buffer.  This program returns "count" strings as
    #   "a", "bb", "ccc", "dddd", et.
    def test_out_strings(self):
        result = ts.out_strings(10)
        (dim1, dim2), [strings] = result
        self.assertEqual((50, 256), (dim1, dim2))  # This is part of the declaration
        self.assertEqual(10, len(strings))
        self.assertEqual("a", strings[0])
        self.assertEqual("j" * 10, strings[9])


class test_multiple_strings_inout(unittest.TestCase):
    # %apply(int DIM1, int DIM2, Type *INOUT_STRINGS)
    # because we could, we wrote a simple sorting program
    def test_basic_test_tuple(self):
        argument = "Four score and thirty years ago".split(' ')
        result, = ts.sort_strings(argument)
        self.assertEqual(sorted(argument), result)

    def test_okay_to_pass_empty_list(self):
        result, = ts.sort_strings(())
        self.assertEqual([], result)

    def test_requires_one_dimensional_array(self):
        argument = np.array([["a", "b", "c", "d"], ["w", "x", "y", "z"]])
        with self.assertRaises(ValueError):
            ts.sort_strings(argument)

    def test_requires_non_null(self):
        with self.assertRaises(TypeError):
            ts.sort_strings(None)


class test_primitive_return_types(unittest.TestCase):
    def test_return_string(self):
        self.assertEqual("hello", ts.return_string())

    def test_return_boolean(self):
        self.assertTrue(ts.return_boolean(100))
        self.assertFalse(ts.return_boolean(0))

    def test_sigerror(self):
        with self.assertRaises(RuntimeError):
            ts.return_sigerr()

if __name__ == '__main__':
    unittest.main()



"""
cspyce/aliases.py

Alias handler for the cspyce library

Aliases allow the user to associate multiple names or codes with the same
CSPICE body or frame. Aliases can be used for a variety of purposes.

- You can use names and codes interchangeably as input arguments to any cspyce
  function.
- You can use a body name or code in place of a frame name or code, and the
  primary frame associated with that identified body will be used.
- Strings that represent integers are equivalent to the integers themselves.

Most importantly, you can allow multiple names or codes to refer to the same
CPSICE body or frame. For bodies and frames that have multiple names or codes,
calls to a cspyce function will try each option in sequence until it finds one
that works. Options are always tried in the order of they were defined, so
higher-priority names and codes are tried first.

Example 1: Jupiter's moon Dia uses code = 553, but it previously used code
55076. With 55076 defined as an alias for 553, a cspyce call will return
information about Dia under either of its codes.

Example 2: The Earth's rotation is, by default, modeled by frame "IAU_EARTH".
However, "ITRF93" is the name of a much more precise description of Earth's
rotation. If you define "IAU_EARTH" as an alias for "ITRF93", then the cspyce
toolkit will use ITRF93 if it is available, and otherwise IAU_EARTH.

Immediately after a cspyce call involving aliases, you can find out what value
or values were actually used by looking at attributes of the function. For
example, the first input to cspyce function spkez is called "targ" and it
identifies the code of a target being observed. After a call to
  cspyce.spkez(553, ...
the value of cspyce.spkez.targ will be the code actually used, in this case
either 553 or 55076.

Upon importing this module, a new function is defined for every cspyce
function that takes a frame or body as input. The new function has the same
name as the pre-existing cspyce function, but with "_alias" inserted
immediately after the original cspyce name (and before any other suffix such
as "_vector" or "_error").

You can make alias support the default for individual cspyce functions or for
the entire cspyce module by calling:
  cspyce.use_aliases()
These versions can subsequently be disabled as the default by calling:
  cspyce.use_noaliases()

You can also select the specific version of the cspyce function via a function
attribute. For example, after calling
  cspyce.use_aliases('spkez')
the function cspyce.spkez will support aliases. However, calling
  cspyce.spkez.noalias(...)
will use the un-aliased version. You can also use the aliased version
explicitly by calling
  cspyce.spkez.alias(...)

To define a body alias or frame alias, call
  cspyce.define_body_aliases(name_or_code, name_or_code, ...)
  cspyce.define_frame_aliases(name_or_code, name_or_code, ...)
where the arguments are an arbitrary list of codes and names.

To determine the aliases associated with a name or code, call
  cspyce.get_body_aliases(name_or_code)
  cspyce.get_frame_aliases(name_or_code)
where the argument is either a name or a code.
"""

import cspyce
import cspyce.alias_support as support

if not hasattr(cspyce, 'ALIASES_IMPORTED'):

    # Upon import, define all missing alias versions
    support._define_all_alias_versions()

    # Also add new functions directly to the cspyce module
    cspyce.use_aliases = support.use_aliases
    cspyce.use_noaliases = support.use_noaliases
    cspyce.get_body_aliases = support.get_body_aliases
    cspyce.get_frame_aliases = support.get_frame_aliases
    cspyce.define_body_aliases = support.define_body_aliases
    cspyce.define_frame_aliases = support.define_frame_aliases

################################################################################
# Record the fact that this module was imported
################################################################################

cspyce.ALIASES_IMPORTED = True

cspyce.use_aliases()

################################################################################



"""
cspyce/arrays.py

Array handler for the cspyce library

This module implements support for multidimensional arrays by building upon
all of the vectorized functions in the cspyce module. Upon importing this
module, a new function is defined for each "_vector" function, with the same
name except "_array" replacing "_vector".

The key difference is that the "_array" functions follow all of the standard
NumPy rules for array broadcasting. Input arguments to cspyce _array functions
can have arbitrary additional dimensions, as long as all those dimensions
broadcast together properly. The returned arrays will all have the
broadcasted shape.
"""

import cspyce
import cspyce.array_support as support

# Upon import, define all array versions
support._define_all_array_versions()

if not hasattr(cspyce, 'ARRAYS_IMPORTED'):

    # Upon import, define all missing alias versions
    support._define_all_array_versions()

    # Also add a new function directly to the cspyce module
    cspyce.use_arrays = support.use_arrays

################################################################################
# Record the fact that this module was imported
################################################################################

cspyce.ARRAYS_IMPORTED = True

################################################################################
